from typing import Optional, Type, Tuple, List

from torch import nn, Tensor

from torchglyph.nn.activation import Activations
from torchglyph.nn.attention import MultiHeadAttention, att_mask, cas_mask

__all__ = [
    'TransformerFfn',
    'TransformerEncoderLayer',
    'TransformerDecoderLayer',
    'Transformer',
]


class TransformerFfn(nn.Sequential):
    def __init__(self, hidden_size: int, dropout: float,
                 activation: Activations = nn.Tanh,
                 bias: bool = True, *, in_size: int) -> None:
        super(TransformerFfn, self).__init__(
            nn.Linear(in_size, hidden_size, bias=bias),
            activation(),
            nn.Dropout(p=dropout, inplace=True),
            nn.Linear(hidden_size, in_size, bias=bias),
        )

        self.in_size = in_size
        self.activation = activation.__name__
        self.dropout = dropout
        self.bias = bias

    def extra_repr(self) -> str:
        return ', '.join([
            f'{self.in_size}',
            f'dropout={self.dropout}',
            f'bias={self.bias}',
        ])

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}{self.activation}({self.extra_repr()})'


class TransformerEncoderLayer(nn.Module):
    def __init__(self, hidden_size: int = 2048, num_heads: int = 8,
                 att_dropout: float = 0., ffn_dropout: float = 0., bias: bool = True, *,
                 in_size: int) -> None:
        super(TransformerEncoderLayer, self).__init__()

        self.in_size = in_size
        self.hidden_size = hidden_size
        self.num_heads = num_heads
        self.att_dropout = att_dropout
        self.ffn_dropout = ffn_dropout
        self.bias = bias

        self.att = MultiHeadAttention(
            num_heads=num_heads,
            head_dim=in_size // num_heads,
            dropout=att_dropout, bias=bias,
            q_dim=in_size, k_dim=in_size, v_dim=in_size,
        )
        self.ffn = TransformerFfn(
            dropout=ffn_dropout, bias=bias,
            in_size=in_size, hidden_size=hidden_size,
        )

        self.norm1 = nn.LayerNorm(in_size)
        self.norm2 = nn.LayerNorm(in_size)
        self.dropout = nn.Dropout(ffn_dropout)

    def forward(self, src: Tensor, src_mask: Optional[Tensor] = None) -> Tensor:
        """
        Args:
            src: [..., s, x]
            src_mask: [..., (h), (s), (s)]
        Returns:
            [..., s, o]
        """
        src = self.norm1(src + self.dropout(self.att(q=src, k=src, v=src, mask=src_mask)))
        src = self.norm2(src + self.dropout(self.ffn(src)))
        return src


class TransformerDecoderLayer(nn.Module):
    def __init__(self, hidden_size: int = 2048, num_heads: int = 8,
                 att_dropout: float = 0., ffn_dropout: float = 0., bias: bool = True, *,
                 in_size: int) -> None:
        super(TransformerDecoderLayer, self).__init__()

        self.in_size = in_size
        self.hidden_size = hidden_size
        self.num_heads = num_heads
        self.att_dropout = att_dropout
        self.ffn_dropout = ffn_dropout
        self.bias = bias

        self.tgt = MultiHeadAttention(
            num_heads=num_heads,
            head_dim=in_size // num_heads,
            dropout=att_dropout, bias=bias,
            q_dim=in_size, k_dim=in_size, v_dim=in_size,
        )
        self.src = MultiHeadAttention(
            num_heads=num_heads,
            head_dim=in_size // num_heads,
            dropout=att_dropout, bias=bias,
            q_dim=in_size, k_dim=in_size, v_dim=in_size,
        )
        self.ffn = TransformerFfn(
            dropout=ffn_dropout, bias=bias,
            in_size=in_size, hidden_size=hidden_size,
        )

        self.norm1 = nn.LayerNorm(in_size)
        self.norm2 = nn.LayerNorm(in_size)
        self.norm3 = nn.LayerNorm(in_size)
        self.dropout = nn.Dropout(ffn_dropout)

    def forward(self, src: Tensor, tgt: Tensor,
                src_mask: Optional[Tensor] = None,
                tgt_mask: Optional[Tensor] = None) -> Tensor:
        """
        Args:
            src: [..., s, x]
            tgt: [..., t, y]
            src_mask: [..., (h), (t), (s)]
            tgt_mask: [..., (h), (t), (t)]
        Returns:
            [..., s, o]
        """

        tgt = self.norm1(tgt + self.dropout(self.tgt(q=tgt, k=tgt, v=tgt, mask=tgt_mask)))
        tgt = self.norm2(tgt + self.dropout(self.src(q=tgt, k=src, v=src, mask=src_mask)))
        tgt = self.norm3(tgt + self.dropout(self.ffn(tgt)))
        return tgt

    def decode(self, tgt: Tensor,
               src_k: Tensor, src_v: Tensor,
               tgt_k: Tensor, tgt_v: Tensor,
               src_mask: Optional[Tensor] = None) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor]:
        tgt_q, tgt_k, tgt_v = self.tgt.decode_tgt(q=tgt, k=tgt_k, v=tgt_v)
        tgt = self.norm1(tgt + self.dropout(tgt_q))

        tgt_q, src_k, src_v = self.src.decode_src(q=tgt, k=src_k, v=src_v, src_mask=src_mask)
        tgt = self.norm2(tgt + self.dropout(tgt_q))

        tgt = self.norm3(tgt + self.dropout(self.ffn(tgt)))
        return tgt, src_k, src_v, tgt_k, tgt_v


class Transformer(nn.Module):
    def __init__(self, num_enc_layers: int = 6, num_dec_layers: int = 6,
                 enc_layer_: Type[TransformerEncoderLayer] = TransformerEncoderLayer,
                 dec_layer_: Type[TransformerDecoderLayer] = TransformerDecoderLayer, *,
                 in_size: int) -> None:
        super(Transformer, self).__init__()

        self.encoder_layers = nn.ModuleList(modules=[
            enc_layer_(in_size=in_size) for _ in range(num_enc_layers)
        ])
        self.decoder_layers = nn.ModuleList(modules=[
            dec_layer_(in_size=in_size) for _ in range(num_dec_layers)
        ])

    def encode(self, src: Tensor, src_mask: Optional[Tensor] = None) -> Tuple[Tensor, Tensor]:
        src_mask = att_mask(mask=src_mask)
        for encoder_layer in self.encoder_layers:
            src = encoder_layer(src=src, src_mask=src_mask)

        return src, src_mask

    def forward(self, src: Tensor, tgt: Tensor,
                src_mask: Optional[Tensor] = None,
                tgt_mask: Optional[Tensor] = None) -> Tensor:
        src, src_mask = self.encode(src=src, src_mask=src_mask)

        tgt_mask = cas_mask(mask=tgt_mask, tensor=tgt, dim=-2)
        for decoder_layer in self.decoder_layers:  # type: TransformerDecoderLayer
            tgt = decoder_layer.forward(
                src=src, src_mask=src_mask,
                tgt=tgt, tgt_mask=tgt_mask,
            )

        return tgt

    Prev = Tuple[List[Tensor], List[Tensor], List[Tensor], List[Tensor]]

    def init_decoding(self, bos: Tensor, src: Tensor, src_mask: Optional[Tensor] = None):
        src, src_mask = self.encode(src=src, src_mask=src_mask)

        src_ks = [src for _ in self.decoder_layers]
        src_vs = [src for _ in self.decoder_layers]
        tgt_ks = [bos for _ in self.decoder_layers]
        tgt_vs = [bos for _ in self.decoder_layers]

        return (src_ks, src_vs, tgt_ks, tgt_vs), src_mask

    def decode(self, tgt: Tensor, prev: Prev, src_mask: Optional[Tensor] = None) -> Tuple[Tensor, Prev]:
        prev_src_ks, prev_src_vs, prev_tgt_ks, prev_tgt_vs = prev

        src_ks, src_vs, tgt_ks, tgt_vs = [], [], [], []
        for index, decoder_layer in enumerate(self.decoder_layers):  # type: (int, TransformerDecoderLayer)
            tgt, src_k_i, src_v_i, tgt_k_i, tgt_v_i = decoder_layer.decode(
                src_k=prev_src_ks[index], src_v=prev_src_vs[index], src_mask=src_mask,
                tgt_k=prev_tgt_ks[index], tgt_v=prev_tgt_vs[index], tgt=tgt,
            )
            src_ks.append(src_k_i)
            src_vs.append(src_v_i)
            tgt_ks.append(tgt_k_i)
            tgt_vs.append(tgt_v_i)

        return tgt, (src_ks, src_vs, tgt_ks, tgt_vs)



from abc import ABCMeta
from datetime import datetime
from logging import getLogger
from typing import Any, Dict

logger = getLogger(__name__)

__all__ = [
    'Meter',
    'AverageMeter', 'AccuracyMeter',
    'ClassificationMeter',
    'TimeMeter',
]


def join(*names: str, sep: str = '.') -> str:
    return sep.join(name for name in names if len(name) > 0)


class Meter(object, metaclass=ABCMeta):
    def __init__(self) -> None:
        super(Meter, self).__init__()
        self.reset_buffers()

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.extra_repr()})'

    def extra_repr(self) -> str:
        return ''

    def reset_buffers(self) -> None:
        raise NotImplementedError

    def update(self, *args, **kwargs) -> None:
        raise NotImplementedError

    @property
    def merit(self) -> float:
        raise NotImplementedError

    def __eq__(self, other: 'Meter') -> bool:
        return self.merit == other.merit

    def __le__(self, other: 'Meter') -> bool:
        return self.merit < other.merit

    def __gt__(self, other: 'Meter') -> bool:
        return self.merit > other.merit

    def __add__(self, other: 'Meter') -> 'Meter':
        raise NotImplementedError

    def state_dict(self, prefix: str, *, destination: Dict[str, Any] = None) -> Dict[str, Any]:
        raise NotImplementedError

    def flush(self, prefix: str, reset_buffers: bool = True) -> None:
        for key, value in self.state_dict(prefix=prefix, destination=None):
            logger.info(f'{key} => {value}')

        if reset_buffers:
            self.reset_buffers()


class AverageMeter(Meter):
    def __init__(self, num_digits: int = 6) -> None:
        super(AverageMeter, self).__init__()
        self.num_digits = num_digits

    def extra_repr(self) -> str:
        return f'num_digits={self.num_digits}'

    def reset_buffers(self) -> None:
        self.total = 0.
        self.weight = 0.

    def update(self, value: float, weight: float) -> None:
        self.total += value
        self.weight += weight

    def __add__(self, other: 'AverageMeter') -> 'AverageMeter':
        meter = self.__class__(num_digits=self.num_digits)
        meter.update(value=self.total, weight=self.weight)
        meter.update(value=other.total, weight=other.weight)
        return meter

    @property
    def average(self) -> float:
        try:
            return round(self.total / self.weight, ndigits=self.num_digits)
        except ZeroDivisionError:
            logger.warning(f'{self.weight} is zero')
            return 0.

    @property
    def merit(self) -> float:
        return self.average

    def state_dict(self, prefix: str, *, destination: Dict[str, Any] = None) -> Dict[str, Any]:
        if destination is None:
            destination = {}

        destination[prefix] = self.merit
        return destination


class AccuracyMeter(AverageMeter):
    def __init__(self, num_digits: int = 2) -> None:
        super(AccuracyMeter, self).__init__(num_digits=num_digits)

    @property
    def average(self) -> float:
        try:
            return round(self.total * 100 / self.weight, ndigits=self.num_digits)
        except ZeroDivisionError:
            logger.warning(f'{self.weight} is zero')
            return 0.


class ClassificationMeter(Meter):
    def __init__(self, num_digits: int = 2) -> None:
        super(ClassificationMeter, self).__init__()
        self.num_digits = num_digits

    def extra_repr(self) -> str:
        return f'num_digits={self.num_digits}'

    def reset_buffers(self) -> None:
        self.total = 0.
        self.prediction_weight = 0.
        self.target_weight = 0.

    def update(self, value: float, prediction_weight: float, target_weight: float) -> None:
        self.total += value
        self.prediction_weight += prediction_weight
        self.target_weight += target_weight

    def __add__(self, other: 'ClassificationMeter') -> 'ClassificationMeter':
        meter = self.__class__(num_digits=self.num_digits)
        meter.update(value=self.total, prediction_weight=self.prediction_weight, target_weight=self.target_weight)
        meter.update(value=other.total, prediction_weight=other.prediction_weight, target_weight=other.target_weight)
        return meter

    @property
    def precision(self) -> float:
        try:
            return round(self.total * 100 / self.prediction_weight, ndigits=self.num_digits)
        except ZeroDivisionError:
            logger.warning(f'{self.prediction_weight} is zero')
            return 0.

    @property
    def recall(self) -> float:
        try:
            return round(self.total * 100 / self.target_weight, ndigits=self.num_digits)
        except ZeroDivisionError:
            logger.warning(f'{self.target_weight} is zero')
            return 0.

    @property
    def f1(self) -> float:
        try:
            return round(self.total * 200 / (self.prediction_weight + self.target_weight), ndigits=self.num_digits)
        except ZeroDivisionError:
            logger.warning(f'{self.prediction_weight + self.target_weight} is zero')
            return 0.

    @property
    def merit(self) -> float:
        return self.f1

    def state_dict(self, prefix: str, *, destination: Dict[str, Any] = None) -> Dict[str, Any]:
        if destination is None:
            destination = {}

        destination[join(prefix, 'precision')] = self.precision
        destination[join(prefix, 'recall')] = self.recall
        destination[join(prefix, 'f1')] = self.f1
        return destination


class TimeMeter(Meter):
    def __init__(self, num_digits: int = 6) -> None:
        super(TimeMeter, self).__init__()
        self.num_digits = num_digits

    def reset_buffers(self) -> None:
        self.seconds = 0
        self.num_units = 0

    def update(self, seconds: float, num_units: float) -> None:
        self.seconds += seconds
        self.num_units += num_units

    def __add__(self, other: 'TimeMeter') -> 'TimeMeter':
        meter = self.__class__(num_digits=self.num_digits)
        meter.update(seconds=self.seconds, num_units=self.num_units)
        meter.update(seconds=other.seconds, num_units=other.num_units)
        return meter

    def tik(self) -> None:
        self.start_datetime = datetime.now()

    def tok(self, num_units: float = 1.) -> None:
        self.seconds += (datetime.now() - self.start_datetime).total_seconds()
        self.num_units += num_units
        del self.start_datetime

    def __enter__(self):
        self.tik()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.tok()

    @property
    def units_per_second(self) -> float:
        try:
            return round(self.num_units / self.seconds, ndigits=self.num_digits)
        except ZeroDivisionError:
            logger.info(f'{self.seconds} is zero')
            return 0.

    @property
    def seconds_per_unit(self) -> float:
        try:
            return round(self.seconds / self.num_units, ndigits=self.num_digits)
        except ZeroDivisionError:
            logger.info(f'{self.num_units} is zero')
            return 0.

    @property
    def merit(self) -> float:
        return self.units_per_second

    def state_dict(self, prefix: str, *, destination: Dict[str, Any] = None) -> Dict[str, Any]:
        if destination is None:
            destination = {}

        destination[join(prefix, 'units_per_second')] = self.units_per_second
        destination[join(prefix, 'seconds_per_unit')] = self.seconds_per_unit
        return destination



import itertools
from collections import namedtuple, OrderedDict
from pathlib import Path
from typing import Iterable, Any, Type
from typing import Union, List, Tuple, NamedTuple, Dict

from torch.distributions.utils import lazy_property
from torch.utils.data import DataLoader as TorchDataLoader, SequentialSampler, RandomSampler
from torch.utils.data import Dataset as TorchDataset
from tqdm import tqdm

from torchglyph.io import DownloadMixin
from torchglyph.pipe import Pipe
from torchglyph.sampler import SizedBatchSampler

__all__ = [
    'Dataset',
    'DataLoader',
]


class Dataset(TorchDataset, DownloadMixin):
    def __init__(self, pipes: List[Dict[str, Pipe]], **kwargs) -> None:
        super(Dataset, self).__init__()

        self.pipes = {}
        self.names = []

        for ps in pipes:
            for name, pipe in ps.items():
                self.pipes[name] = pipe
                self.names.append(name)

        self.data = {}
        for datum, ps in zip(zip(*self.load(**kwargs)), pipes):
            for name, pipe in ps.items():
                self.data.setdefault(name, []).extend(datum)

    def get_size(self, item: Any) -> int:
        raise NotImplementedError

    def __getitem__(self, index: int) -> Dict[str, Any]:
        return {name: self.data[name][index] for name in self.names}

    def __len__(self) -> int:
        return len(next(iter(self.data.values())))

    @lazy_property
    def named_tuple(self) -> Type:
        return namedtuple(f'{self.__class__.__name__}Batch', field_names=self.names)

    @property
    def vocabs(self) -> NamedTuple:
        return self.named_tuple(**{
            name: pipe.vocab
            for name, pipe in self.pipes.items()
        })

    def collate_fn(self, batch: List[Dict[str, Any]]) -> NamedTuple:
        return self.named_tuple(**{
            name: pipe.collate_fn([data[name] for data in batch])
            for name, pipe in self.pipes.items()
        })

    @classmethod
    def load(cls, **kwargs) -> Iterable[Any]:
        raise NotImplementedError

    def dump(self, fp, batch: NamedTuple, prediction: Any, *args, **kwargs) -> None:
        raise NotImplementedError

    def state_dict(self, destination: OrderedDict = None, prefix: str = '',
                   keep_vars: bool = False) -> OrderedDict:
        if destination is None:
            destination = OrderedDict()
            destination._metadata = OrderedDict()

        for name, datum in self.data.items():
            destination[prefix + name] = datum

        return destination

    def load_state_dict(self, state_dict: OrderedDict, strict: bool = True) -> None:
        names = set(self.names)
        for name, datum in state_dict.items():
            self.data[name] = datum
            if strict:
                names.remove(name)

        if strict:
            assert len(names) == 0

    def eval(self, path: Path, **kwargs):
        raise NotImplementedError

    @classmethod
    def new(cls, **kwargs) -> Tuple['DataLoader', ...]:
        raise NotImplementedError


class DataLoader(TorchDataLoader):
    dataset: Dataset

    @property
    def vocabs(self) -> NamedTuple:
        return self.dataset.vocabs

    @classmethod
    def new(cls, datasets: Tuple[Dataset, ...],
            batch_size: Union[int, Tuple[int, ...]],
            shuffle: bool = True, drop_last: bool = False) -> List['DataLoader']:
        assert len(datasets) > 0

        batch_sizes = batch_size
        if isinstance(batch_size, int):
            batch_sizes = itertools.repeat(batch_size)

        with tqdm(desc='post-processing', total=sum(len(dataset.pipes) for dataset in datasets)) as progress:
            for index, dataset in enumerate(datasets):
                for name, pipe in dataset.pipes.items():
                    progress.set_postfix_str(f'{index}.{name}')
                    pipe.postprocess_(dataset)
                    progress.update(1)

        loaders = []

        for index, (dataset, batch_size) in enumerate(zip(datasets, batch_sizes)):
            if index == 0 and shuffle:
                sampler = RandomSampler(dataset, generator=None)
            else:
                sampler = SequentialSampler(dataset)

            batch_sampler = SizedBatchSampler(
                dataset=dataset, sampler=sampler, batch_size=batch_size,
                drop_last=index == 0 and drop_last,
            )

            loaders.append(DataLoader(
                dataset=dataset, batch_size=1,
                collate_fn=dataset.collate_fn,
                batch_sampler=batch_sampler,
            ))

        return loaders



from typing import Tuple, Optional

import torch
from einops.layers.torch import Rearrange
from torch import Tensor
from torch import nn

__all__ = [
    'att_mask',
    'cas_mask',
    'MultiHeadAttention',
]


@torch.no_grad()
def att_mask(mask: Optional[Tensor] = None) -> Optional[Tensor]:
    """
    Args:
        mask: [..., k]
    Returns:
        [..., (h), (q), k]
    """
    return mask if mask is None else mask[..., None, None, :]


@torch.no_grad()
def cas_mask(tensor: Tensor, dim: int = -2, mask: Optional[Tensor] = None) -> Optional[Tensor]:
    """
        Args:
            tensor: [..., k, ...]
            dim: []
            mask: [..., k]
        Returns:
            [..., (h), k, k]
        """
    cas = torch.ones((tensor.size()[dim], tensor.size()[dim]), device=tensor.device, dtype=torch.bool).triu(1)
    return cas if mask is None else torch.logical_or(mask[..., None, None, :], cas)


class MultiHeadAttention(nn.Module):
    def __init__(self, num_heads: int = 8, head_dim: int = 64,
                 dropout: float = 0., bias: bool = True, *,
                 q_dim: int, k_dim: int, v_dim: int) -> None:
        super(MultiHeadAttention, self).__init__()

        self.q_dim = q_dim
        self.k_dim = k_dim
        self.v_dim = v_dim
        self.num_heads = num_heads
        self.head_dim = head_dim
        self.dropout = dropout
        self.bias = bias
        self.tau = head_dim ** -0.5

        self.q = nn.Sequential(
            nn.Linear(q_dim, num_heads * head_dim, bias=bias),
            Rearrange('... q (h x) -> ... h q x', h=num_heads),
        )
        self.k = nn.Sequential(
            nn.Linear(k_dim, num_heads * head_dim, bias=bias),
            Rearrange('... k (h x) -> ... h x k', h=num_heads),
        )
        self.softmax = nn.Sequential(
            nn.Dropout(dropout, inplace=True),
            nn.Softmax(dim=-1),
        )

        self.v = nn.Sequential(
            nn.Linear(v_dim, num_heads * head_dim, bias=bias),
            Rearrange('... k (h x) -> ... h k x', h=num_heads),
        )
        self.o = nn.Sequential(
            Rearrange('... h q x -> ... q (h x)'),
            nn.Linear(num_heads * head_dim, q_dim, bias=bias),
        )

    def extra_repr(self) -> str:
        return ', '.join([
            f'q={self.q_dim}', f'k={self.k_dim}', f'v={self.v_dim}',
            f'heads={self.head_dim}(x{self.num_heads})',
            f'dropout={self.dropout}', f'bias={self.bias}',
        ])

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.extra_repr()})'

    def forward(self, q: Tensor, k: Tensor, v: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        """
        Args:
            q: [..., q, x]
            k: [..., k, y]
            v: [..., k, z]
            mask: [..., (h), (q), k]
        Returns:
            [..., q, o]
        """

        k = self.k(k)
        v = self.v(v)
        q = self.q(q)

        attention = q @ k * self.tau
        if mask is not None:
            attention, mask = torch.broadcast_tensors(attention, mask)
            attention.masked_fill_(mask=mask, value=-float('inf'))
        attention = self.softmax(attention)

        return self.o(attention @ v)

    def decode_tgt(self, q: Tensor, k: Tensor, v: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
        """
        Args:
            q: [..., q, x]
            k: [..., k, y] or [..., h, t, k]
            v: [..., k, z] or [..., h, k, t]
        Returns:
            [..., q, o], [..., h, t, k + 1], [..., h, k + 1, t]
        """

        k = self.k(k) if q.dim() == k.dim() else torch.cat([k, self.k(q)], dim=-1)
        v = self.v(v) if q.dim() == v.dim() else torch.cat([v, self.v(q)], dim=-2)
        q = self.q(q)

        attention = self.softmax(q @ k * self.tau)

        return self.o(attention @ v), k, v

    def decode_src(self, q: Tensor, k: Tensor, v: Tensor,
                   src_mask: Optional[Tensor] = None) -> Tuple[Tensor, Tensor, Tensor]:
        """
        Args:
            q: [..., q, x]
            k: [..., k, y] or [..., h, s, k]
            v: [..., k, z] or [..., h, k, s]
            src_mask: [..., (h), (q), k]
        Returns:
            [..., q, o], [..., h, s, k], [..., h, k, s]
        """

        k = self.k(k) if q.dim() == k.dim() else k
        v = self.v(v) if q.dim() == v.dim() else v
        q = self.q(q)

        attention = q @ k * self.tau

        if src_mask is not None:
            attention, src_mask = torch.broadcast_tensors(attention, src_mask)
            attention.masked_fill_(mask=src_mask, value=-float('inf'))
        attention = self.softmax(attention)

        return self.o(attention @ v), k, v



from abc import ABCMeta, abstractmethod
from typing import Optional, Union, Any, List, Set, Tuple

__all__ = [
    'compress', 'subs',
    'Proc', 'Processors',
    'Identity', 'Lift', 'Chain',
    'Map', 'Filter',
]

Processors = Union[Optional['Proc'], List[Optional['Proc']]]


def compress(processors: Processors, allow_ellipsis: bool = True) -> List['Proc']:
    if processors is None or isinstance(processors, Identity):
        return []
    if processors is ...:
        if allow_ellipsis:
            return [...]
        else:
            raise ValueError(f'ellipsis is not allowed here, {processors}')
    if isinstance(processors, Chain):
        return processors.proc
    if isinstance(processors, Proc):
        return [processors]
    return [p for proc in processors for p in compress(proc, allow_ellipsis=allow_ellipsis)]


def subs(processors: Processors, repl: Processors) -> Processors:
    return [repl if proc is ... else proc for proc in compress(processors, allow_ellipsis=True)]


class Proc(object, metaclass=ABCMeta):
    @classmethod
    def from_list(cls, processors: List['Proc']) -> 'Proc':
        if len(processors) == 0:
            return Identity()
        if len(processors) == 1:
            return processors[0]
        return Chain(processors)

    def extra_repr(self) -> str:
        return f''

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.extra_repr()})'

    def __add__(self, processors: Processors) -> 'Proc':
        return self.from_list([self] + compress(processors))

    def __radd__(self, processors: Processors) -> 'Proc':
        return self.from_list(compress(processors) + [self])

    @abstractmethod
    def __call__(self, data: Any, **kwargs) -> Any:
        raise NotImplementedError


class Identity(Proc):
    def __repr__(self) -> str:
        return f'{None}'

    def __call__(self, data: Any, **kwargs) -> Any:
        return data


class Lift(Proc):
    Sequence = Union[Set[Any], List[Any], Tuple[Any, ...]]

    def __init__(self, proc: Proc) -> None:
        super(Lift, self).__init__()
        self.proc = proc

    def __repr__(self) -> str:
        return f'[{self.proc.__repr__()}]'

    def __call__(self, sequence: Sequence, **kwargs) -> Sequence:
        return type(sequence)([self.proc(item, **kwargs) for item in sequence])


class Chain(Proc):
    def __init__(self, processors: Processors) -> None:
        super(Chain, self).__init__()
        self.proc = compress(processors)

    def extra_repr(self) -> str:
        return ' + '.join([str(proc) for proc in self.proc])

    def __repr__(self) -> str:
        return self.extra_repr()

    def __add__(self, processors: Processors) -> 'Proc':
        return self.from_list(self.proc + compress(processors))

    def __radd__(self, processors: Processors) -> 'Proc':
        return self.from_list(compress(processors) + self.proc)

    def __call__(self, data: Any, **kwargs) -> Any:
        for proc in self.proc:
            data = proc(data, **kwargs)
        return data


class Map(Proc):
    Sequence = Union[Any, Set[Any], List[Any], Tuple[Any, ...]]

    def map(self, data: Any, **kwargs) -> Any:
        raise NotImplementedError

    def __call__(self, sequence: Sequence, **kwargs) -> Sequence:
        if not isinstance(sequence, (set, list, tuple)):
            return self.map(sequence, **kwargs)
        return type(sequence)([self(data, **kwargs) for data in sequence])


class Filter(Proc):
    Sequence = Union[Any, Set[Any], List[Any], Tuple[Any, ...]]

    def filter(self, data: Any, **kwargs) -> bool:
        raise NotImplementedError

    def __call__(self, sequence: Sequence, **kwargs) -> Sequence:
        return type(sequence)([data for data in sequence if self.filter(data, **kwargs)])



from abc import ABCMeta
from typing import List, Any

from torch import Tensor
from torch.nn.utils.rnn import PackedSequence
from torch.types import Device, Number
from torchrua import pad_sequence, pad_packed_sequence, pad_catted_sequence, CattedSequence

from torchglyph.proc.abc import Proc

__all__ = [
    'PaddingProc',
    'PadSequence',
    'PadCattedSequence',
    'PadPackedSequence',
]


class PaddingProc(Proc, metaclass=ABCMeta):
    def __init__(self, batch_first: bool = True,
                 padding_value: Number = 0, device: Device = None) -> None:
        super(PaddingProc, self).__init__()
        self.batch_first = batch_first
        self.padding_value = padding_value
        self.device = device

    def extra_repr(self) -> str:
        return ', '.join([
            f'batch_first={self.batch_first}',
            f'padding_value={self.padding_value}',
            f'device={self.device}',
        ])

    def __call__(self, data: Any, **kwargs) -> Tensor:
        raise NotImplementedError


class PadSequence(PaddingProc):
    def __call__(self, data: List[Tensor], **kwargs) -> Tensor:
        return pad_sequence(
            sequences=data, batch_first=self.batch_first,
            padding_value=self.padding_value, device=self.device,
        )


class PadPackedSequence(PaddingProc):
    def __call__(self, data: PackedSequence, **kwargs) -> Tensor:
        data, _ = pad_packed_sequence(
            sequence=data, batch_first=self.batch_first,
            padding_value=self.padding_value, device=self.device,
        )
        return data


class PadCattedSequence(PaddingProc):
    def __call__(self, data: CattedSequence, **kwargs) -> Tensor:
        data, token_sizes = data
        return pad_catted_sequence(
            sequence=data, token_sizes=token_sizes, batch_first=self.batch_first,
            padding_value=self.padding_value, device=self.device,
        )



import logging
from pathlib import Path
from typing import Iterable, Any
from typing import Tuple

import torch
from torch.types import Device
from tqdm import tqdm

from torchglyph import data_dir
from torchglyph.dataset import Dataset, DataLoader
from torchglyph.pipe import PadListStrPipe

__all__ = [
    'MachineTranslation',
    'IWSLT14',
]


class WordPipe(PadListStrPipe):
    def __init__(self, device: Device) -> None:
        super(WordPipe, self).__init__(
            batch_first=True, device=device,
            unk_token='<unk>', pad_token='<pad>',
            special_tokens=('<bos>', '<eos>'),
            threshold=16, dtype=torch.long,
        )


class MachineTranslation(Dataset):
    @classmethod
    def load(cls, path: Path, src_lang: str, tgt_lang: str, encoding: str = 'utf-8', **kwargs) -> Iterable[Any]:
        src_path = path.with_name(f'{path.name}.{src_lang}')
        tgt_path = path.with_name(f'{path.name}.{tgt_lang}')
        with src_path.open(mode='r', encoding=encoding) as src_fp:
            with tgt_path.open(mode='r', encoding=encoding) as tgt_fp:
                for src, tgt in tqdm(zip(src_fp, tgt_fp), desc=f'{path.resolve()}'):
                    yield [src.strip().split(' '), tgt.strip().split(' ')]

    @classmethod
    def new(cls, batch_size: int, share_vocab: bool, src_lang: str, tgt_lang: str, *,
            device: Device, root: Path = data_dir, **kwargs) -> Tuple['DataLoader', ...]:
        if share_vocab:
            src = tgt = WordPipe(device=device)
        else:
            src = WordPipe(device=device)
            tgt = WordPipe(device=device)

        pipes = [
            dict(src=src),
            dict(tgt=tgt),
        ]

        for ps in pipes:
            for name, pipe in ps.items():
                logging.info(f'{name} => {pipe}')

        train, dev, test = cls.paths(root=root)

        train = cls(pipes=pipes, path=train, src_lang=src_lang, tgt_lang=tgt_lang)
        dev = cls(pipes=pipes, path=dev, src_lang=src_lang, tgt_lang=tgt_lang)
        test = cls(pipes=pipes, path=test, src_lang=src_lang, tgt_lang=tgt_lang)

        src.build_vocab_(train)
        if not share_vocab:
            tgt.build_vocab_(train)

        return DataLoader.new(
            (train, dev, test),
            batch_size=batch_size,
            shuffle=True, drop_last=False,
        )


class IWSLT14(MachineTranslation):
    urls = [(
        'https://raw.githubusercontent.com/pytorch/fairseq/master/examples/translation/prepare-iwslt14.sh',
        'prepare-iwslt14.sh',
    )]

    @classmethod
    def paths(cls, root: Path = data_dir, **kwargs) -> Tuple[Path, ...]:
        path, = super(IWSLT14, cls).paths(root=root, **kwargs)
        train = path.parent / 'iwslt14.tokenized.de-en' / 'train'
        dev = path.parent / 'iwslt14.tokenized.de-en' / 'valid'
        test = path.parent / 'iwslt14.tokenized.de-en' / 'test'
        return train, dev, test


if __name__ == '__main__':
    train, dev, test = IWSLT14.new(
        batch_size=32, share_vocab=False,
        src_lang='en', tgt_lang='de',
        device=torch.device('cpu'),
    )
    for item in train:
        print(item)



from typing import Optional, Tuple, List

import torch
from torch.nn.utils.rnn import PackedSequence
from torch.types import Device, Number
from torchrua.padding import pad_packed_sequence

from torchglyph.pipe.abc import Pipe
from torchglyph.proc.abc import Lift
from torchglyph.proc.basic import ToList
from torchglyph.proc.catting import CatSequence
from torchglyph.proc.collating import ToTensor
from torchglyph.proc.packing import PackSequence, ReduceCattedSequences
from torchglyph.proc.vocab import UpdateCounter, BuildVocab, StatsVocab, Numbering

__all__ = [
    'PackListNumPipe', 'PackListListNumPipe',
    'PackListStrPipe', 'PackListListStrPipe',
]


class PackListNumPipe(Pipe):
    def __init__(self, device: Device, dtype: torch.dtype = torch.long) -> None:
        super(PackListNumPipe, self).__init__()
        self.with_(
            post=ToTensor(dtype=dtype),
            batch=PackSequence(device=device),
        )

    def inv(self, sequence: PackedSequence) -> List[List[Number]]:
        data, token_sizes = pad_packed_sequence(sequence=sequence, batch_first=True)

        data = data.detach().cpu().tolist()
        token_sizes = token_sizes.detach().cpu().tolist()

        return [
            [data[index1][index2] for index2 in range(token_size)]
            for index1, token_size in enumerate(token_sizes)
        ]


class PackListStrPipe(PackListNumPipe):
    def __init__(self, device: Device,
                 unk_token: Optional[str], special_tokens: Tuple[Optional[str], ...] = (),
                 threshold: int = 10, dtype: torch.dtype = torch.long) -> None:
        super(PackListStrPipe, self).__init__(device=device, dtype=dtype)
        self.with_(
            pre=UpdateCounter(),
            vocab=[
                BuildVocab(unk_token=unk_token, pad_token=None, special_tokens=special_tokens),
                StatsVocab(threshold=threshold),
            ],
            post=Numbering() + ...,
        )

    def inv(self, sequence: PackedSequence) -> List[List[str]]:
        assert sequence.data.dim() == 1, f'{sequence.data.dim()} != 1'

        return [
            [self.vocab.inv(index) for index in indices]
            for indices in super(PackListStrPipe, self).inv(sequence)
        ]


class PackListListNumPipe(Pipe):
    def __init__(self, device: Device, dtype: torch.dtype = torch.long) -> None:
        super(PackListListNumPipe, self).__init__(
            pre=None,
            vocab=None,
            post=Lift(ToTensor(dtype=dtype)) + CatSequence(device=None),
            batch=ReduceCattedSequences(device=device),
        )


class PackListListStrPipe(PackListListNumPipe):
    def __init__(self, device: Device,
                 unk_token: Optional[str], special_tokens: Tuple[Optional[str], ...] = (),
                 threshold: int = 10, dtype: torch.dtype = torch.long) -> None:
        super(PackListListStrPipe, self).__init__(device=device, dtype=dtype)
        self.with_(
            pre=Lift(ToList() + UpdateCounter()),
            vocab=[
                BuildVocab(unk_token=unk_token, pad_token=None, special_tokens=special_tokens),
                StatsVocab(threshold=threshold),
            ],
            post=Lift(Numbering()) + ...,
        )



import torch
from torch import Tensor
from torch.nn import init

from torch.nn.init import calculate_gain, constant_, orthogonal_

__all__ = [
    'xavier_normal_', 'xavier_uniform_',
    'kaiming_normal_', 'kaiming_uniform_',
    'bert_normal_',
    'constant_', 'orthogonal_',
]


@torch.no_grad()
def xavier_normal_(tensor: Tensor, fan_in: int, fan_out: int, gain: float = 1.) -> Tensor:
    std = gain * (2.0 / (fan_in + fan_out)) ** 0.5
    return init.normal_(tensor, mean=0, std=std)


@torch.no_grad()
def xavier_uniform_(tensor: Tensor, fan_in: int, fan_out: int, gain: float = 1.) -> Tensor:
    bound = gain * (6.0 / (fan_in + fan_out)) ** 0.5
    return init.uniform_(tensor, a=-bound, b=+bound)


@torch.no_grad()
def kaiming_normal_(tensor: Tensor, fan: int, a: float = 0., nonlinearity: str = 'leaky_relu') -> Tensor:
    std = calculate_gain(nonlinearity, a) / fan ** 0.5
    return init.normal_(tensor, mean=0, std=std)


@torch.no_grad()
def kaiming_uniform_(tensor: Tensor, fan: int, a: float = 0., nonlinearity: str = 'leaky_relu') -> Tensor:
    bound = calculate_gain(nonlinearity, a) * (3.0 / fan) ** 0.5
    return init.uniform_(tensor, a=-bound, b=+bound)


@torch.no_grad()
def bert_normal_(tensor: Tensor, mean: float = 0., std: float = 0.02) -> Tensor:
    return init.normal_(tensor, mean=mean, std=std)



from torch import nn, Tensor
from torch.nn import GLU, Tanh, ReLU, GELU, SELU

__all__ = [
    'GLU', 'GTU',
    'Tanh', 'ReLU', 'GELU', 'SELU',
]


class GTU(nn.Module):
    def __init__(self, dim: int = -1) -> None:
        super(GTU, self).__init__()
        self.dim = dim

    def extra_repr(self) -> str:
        return f'dim={self.dim}'

    def forward(self, tensor: Tensor) -> Tensor:
        a, b = tensor.chunk(2, dim=self.dim)
        return a.tanh_() * b.sigmoid_()



from typing import Optional, Tuple, List

import torch
from torch import Tensor
from torch.types import Device, Number

from torchglyph.pipe.abc import Pipe
from torchglyph.proc.collating import ToTensor, ToDevice
from torchglyph.proc.padding import PadSequence
from torchglyph.proc.vocab import UpdateCounter, BuildVocab, StatsVocab, Numbering

__all__ = [
    'PadNumPipe', 'PadListNumPipe',
    'PadStrPipe', 'PadListStrPipe',
]


class PadNumPipe(Pipe):
    def __init__(self, device: Device, dtype: torch.dtype = torch.long) -> None:
        super(PadNumPipe, self).__init__(
            pre=None,
            vocab=None,
            post=None,
            batch=ToTensor(dtype=dtype) + ToDevice(device=device),
        )

    def inv(self, data: Tensor) -> List[Number]:
        return data.detach().cpu().tolist()


class PadStrPipe(PadNumPipe):
    def __init__(self, device: Device, unk_token: Optional[str] = None,
                 pad_token: str = '<pad>', special_tokens: Tuple[Optional[str], ...] = (),
                 threshold: int = 10, dtype: torch.dtype = torch.long) -> None:
        super(PadStrPipe, self).__init__(device=device, dtype=dtype)
        self.with_(
            pre=UpdateCounter(),
            vocab=[
                BuildVocab(unk_token=unk_token, pad_token=pad_token, special_tokens=special_tokens),
                StatsVocab(threshold=threshold),
            ],
            post=Numbering() + ...,
            batch=...,
        )

    def inv(self, data: Tensor) -> List[str]:
        return [
            self.vocab.inv(index)
            for index in super(PadStrPipe, self).inv(data=data)
        ]


class PadListNumPipe(Pipe):
    def __init__(self, batch_first: bool, padding_value: Number,
                 device: Device, dtype: torch.dtype = torch.long) -> None:
        super(PadListNumPipe, self).__init__(
            post=ToTensor(dtype=dtype),
            batch=PadSequence(batch_first=batch_first, padding_value=padding_value, device=device),
        )

    def inv(self, data: Tensor, token_sizes: Tensor) -> List[List[Number]]:
        data = data.detach().cpu().tolist()
        token_sizes = token_sizes.detach().cpu().tolist()

        return [
            [data[index1][index2] for index2 in range(token_size)]
            for index1, token_size in enumerate(token_sizes)
        ]


class PadListStrPipe(PadListNumPipe):
    def __init__(self, batch_first: bool, device: Device,
                 unk_token: Optional[str], pad_token: str = '<pad>',
                 special_tokens: Tuple[Optional[str], ...] = (),
                 threshold: int = 10, dtype: torch.dtype = torch.long) -> None:
        super(PadListStrPipe, self).__init__(
            batch_first=batch_first,
            padding_value=0,  # TODO: fix padding_value
            device=device, dtype=dtype,
        )
        self.with_(
            pre=UpdateCounter(),
            vocab=[
                BuildVocab(unk_token=unk_token, pad_token=pad_token, special_tokens=special_tokens),
                StatsVocab(threshold=threshold),
            ],
            post=Numbering() + ...,
        )

    def inv(self, data: Tensor, token_sizes: Tensor) -> List[List[str]]:
        assert data.dim() == 2, f'{data.dim()} != 2'
        assert token_sizes.dim() == 1, f'{token_sizes.dim()} == {1}'

        return [
            [self.vocab.inv(index) for index in indices]
            for indices in super(PadListStrPipe, self).inv(data=data, token_sizes=token_sizes)
        ]



from torchglyph.proc.abc import *
from torchglyph.proc.basic import *
from torchglyph.proc.catting import *
from torchglyph.proc.collating import *
from torchglyph.proc.ctx import *
from torchglyph.proc.packing import *
from torchglyph.proc.padding import *
from torchglyph.proc.vocab import *



from typing import Union

import torch
from einops import rearrange
from torch import Tensor
from torch import nn
from torch.nn.utils.rnn import PackedSequence
from torchrua import PackedMeta, PackedSequential
from torchrua import batch_sizes_to_ptr

from torchglyph.nn.init import bert_normal_
from torchglyph.vocab import Vocab

__all__ = [
    'TokenEmbedding',
    'CharLstmEmbedding',
    'PositionalEmbedding',
    'TriangularEmbedding',
]


class TokenEmbedding(nn.Embedding, metaclass=PackedMeta):
    def __init__(self, embedding_dim: int, freeze: bool = False, *, vocab: Vocab = None,
                 num_embeddings: int = 0, padding_idx: int = None,
                 dtype: torch.dtype = torch.float32) -> None:
        if vocab is not None:
            super(TokenEmbedding, self).__init__(
                embedding_dim=embedding_dim,
                num_embeddings=len(vocab),
                padding_idx=vocab.pad_idx,
                _weight=vocab.vectors, dtype=dtype,
            )
        else:
            super(TokenEmbedding, self).__init__(
                embedding_dim=embedding_dim,
                num_embeddings=num_embeddings,
                padding_idx=padding_idx,
                _weight=None, dtype=dtype,
            )
        self.weight.requires_grad = not freeze

    def extra_repr(self) -> str:
        args = [super(TokenEmbedding, self).extra_repr()]
        if not self.weight.requires_grad:
            args.append('frozen')
        return ', '.join(args)

    @torch.no_grad()
    def padding_mask(self, indices: Tensor) -> Tensor:
        return indices == (self.padding_idx if self.padding_idx is None else -100)


class CharLstmEmbedding(nn.Module):
    def __init__(self, hidden_dim: int = 50, num_layers: int = 1,
                 bias: bool = True, bidirectional: bool = True, dropout: float = 0.5, *,
                 char_embedding: TokenEmbedding, dtype: torch.dtype = torch.float32) -> None:
        super(CharLstmEmbedding, self).__init__()

        self.embedding = PackedSequential(
            char_embedding,
            nn.Dropout(dropout),
        )

        self.rnn = nn.LSTM(
            input_size=char_embedding.embedding_dim,
            hidden_size=hidden_dim,
            num_layers=num_layers, bias=bias,
            bidirectional=bidirectional,
            dtype=dtype,
        )

        self.num_directions = 2 if self.rnn.bidirectional else 1
        self.embedding_dim = self.rnn.hidden_size * self.num_directions

    def forward(self, indices: PackedSequence) -> Tensor:
        embedding = self.embedding(indices)
        _, (encoding, _) = self.rnn(embedding)
        return rearrange(encoding, '(l d) b x -> l b (d x)', d=self.num_directions)[-1]


class PositionalEmbedding(nn.Module):
    def __init__(self, embedding_dim: int, num_embeddings: int = 1024, freeze: bool = False, *,
                 dtype: torch.dtype = torch.float32) -> None:
        super(PositionalEmbedding, self).__init__()

        self.num_embeddings = num_embeddings
        self.embedding_dim = embedding_dim

        self.weight = nn.Parameter(self.obtain_parameters(dtype=dtype), requires_grad=not freeze)

    def extra_repr(self) -> str:
        args = [
            f'{self.num_embeddings}',
            f'{self.embedding_dim}',
        ]
        if not self.weight.requires_grad:
            args.append('frozen')
        return ', '.join(args)

    @torch.no_grad()
    def obtain_parameters(self, dtype: torch.dtype, **kwargs) -> Tensor:
        return bert_normal_(torch.empty((self.num_embeddings, self.embedding_dim), dtype=dtype))

    def forward(self, indices: Union[Tensor, PackedSequence], dim: int = -2) -> Union[Tensor, PackedSequence]:
        if torch.is_tensor(indices):
            return self.weight[:indices.size()[dim]]

        batch_sizes = indices.batch_sizes.to(device=indices.data.device)
        indices, _, _ = batch_sizes_to_ptr(batch_sizes=batch_sizes)
        return torch.embedding(weight=self.weight, indices=indices)


class TriangularEmbedding(PositionalEmbedding):
    def __init__(self, embedding_dim: int, num_embeddings: int = 1024, freeze: bool = True, *,
                 dtype: torch.dtype = torch.float32) -> None:
        super(TriangularEmbedding, self).__init__(
            embedding_dim=embedding_dim,
            num_embeddings=num_embeddings,
            freeze=freeze, dtype=dtype,
        )

    @torch.no_grad()
    def obtain_parameters(self, dtype: torch.dtype, **kwargs) -> Tensor:
        token = torch.arange(self.num_embeddings, dtype=dtype)
        feature = torch.arange(self.embedding_dim // 2, dtype=dtype) * 2

        position = token[:, None] / (10000. ** (feature[None, :] / self.embedding_dim))
        embedding = torch.stack([torch.sin(position), torch.cos(position)], dim=-1)
        return torch.flatten(embedding, start_dim=-2)



import json
from logging import getLogger
from pathlib import Path
from typing import Any, Set
from typing import List, Tuple, Dict

import torch
from tabulate import tabulate
from torch import nn

logger = getLogger(__name__)

LOG_TXT = 'log.txt'
ARGS_JSON = 'args.json'
SOTA_JSON = 'sota.json'
CHECKPOINT_PT = 'checkpoint.pt'


def save_kwargs(name: str, *, out_dir: Path, **kwargs) -> None:
    if (out_dir / name).exists():
        with (out_dir / name).open(mode='r', encoding='utf-8') as fp:
            kwargs = {**json.load(fp=fp), **kwargs}

    with (out_dir / name).open(mode='w', encoding='utf-8') as fp:
        json.dump(kwargs, fp=fp, indent=2, sort_keys=True)


def save_args(*, out_dir: Path, **kwargs) -> None:
    return save_kwargs(name=ARGS_JSON, out_dir=out_dir, **kwargs)


def save_sota(*, out_dir: Path, **kwargs) -> None:
    return save_kwargs(name=SOTA_JSON, out_dir=out_dir, **kwargs)


def save_checkpoint(name: str = CHECKPOINT_PT, *, out_dir: Path, **kwargs) -> None:
    logger.info(f'saving checkpoint ({", ".join(kwargs.keys())}) to {out_dir / name}')
    return torch.save({name: module.state_dict() for name, module in kwargs.items()}, f=out_dir / name)


def load_checkpoint(name: str = CHECKPOINT_PT, strict: bool = True, *, out_dir: Path, **kwargs) -> None:
    state_dict = torch.load(out_dir / name, map_location=torch.device('cpu'))

    for name, module in kwargs.items():  # type: str, nn.Module
        logger.info(f'loading {name}.checkpoint from {out_dir / name}')
        missing_keys, unexpected_keys = module.load_state_dict(state_dict=state_dict[name], strict=strict)

        if not strict:
            for missing_key in missing_keys:
                logger.warning(f'{name}.{missing_key} is missing')

            for unexpected_key in unexpected_keys:
                logger.warning(f'{name}.{unexpected_key} is unexpected')


def fetch_one(out_dir: Path) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    with (out_dir / SOTA_JSON).open(mode='r', encoding='utf-8') as fp:
        sota = json.load(fp)
    with (out_dir / ARGS_JSON).open(mode='r', encoding='utf-8') as fp:
        args = json.load(fp)

    return {**args, 'path': out_dir / LOG_TXT}, sota


def fetch_all(paths: List[Path]):
    return zip(*[
        fetch_one(out_dir=out_dir)
        for p in paths for out_dir in p.iterdir()
        if out_dir.is_dir() and (out_dir / ARGS_JSON).exists() and (out_dir / SOTA_JSON).exists()
    ])


def group_keys(keys: Set[str], args: List[Dict[str, Any]]):
    has_path = False
    major, common = [], {}

    for key in keys:
        values = set(a.get(key, '-') for a in args)
        if len(values) > 1:
            if key == 'path':
                has_path = True
            else:
                major.append(key)
        else:
            common[key] = args[0][key]

    major = sorted(major)
    if has_path:
        major = major + ['path']

    return major, list(common.items())


def reduce_metric(metrics: List[Tuple[float]], expand: bool):
    metrics = torch.tensor(metrics, dtype=torch.float32)
    s, m = torch.std_mean(metrics[:, 0], unbiased=True)
    epoch1 = round(metrics[:, 1].mean().item())
    epoch2 = round(metrics[:, 2].mean().item())
    if expand:
        return f'{m.item():.4f}', epoch1, epoch2
    return f'{m.item():.4f}', f'{s.item():.2f}', len(metrics), epoch1, epoch2


def summary(path: List[Path], metric: str, sort: bool = True,
            common: bool = False, expand: bool = False, fmt: str = 'pretty'):
    args, sota = fetch_all(path)

    ignores = ('study', 'device', 'seed')
    if not expand:
        ignores = (*ignores, 'path')

    keys = set(k for a in args for k in a.keys() if k not in ignores)
    keys, tabular_data = group_keys(keys=keys, args=args)

    if common:
        print(tabulate(tabular_data=tabular_data, headers=['key', 'value'], tablefmt=fmt))
    else:
        if metric.startswith('dev_'):
            epoch = 'dev_epoch'
        elif metric.startswith('test_'):
            epoch = 'test_epoch'
        else:
            epoch = 'epoch'

        tabular_data = {}
        for s, a in zip(sota, args):
            if metric in s:
                vs = tuple(a.get(key, '-') for key in keys)
                tabular_data.setdefault(vs, []).append((s[metric], s[epoch], s['epoch']))

        tabular_data = [
            [*reduce_metric(metrics, expand), *vs]
            for vs, metrics in tabular_data.items()
        ]
        if sort:
            tabular_data = list(sorted(tabular_data, key=lambda item: item[0], reverse=False))

        if expand:
            headers = [metric, 's', 'e', *keys]
        else:
            headers = [metric, 'std', '*', 's', 'e', *keys]

        print(tabulate(tabular_data=tabular_data, headers=headers, tablefmt=fmt))



from typing import Iterable

from torch import nn, optim

__all__ = [
    'SGD', 'Adam',
]


class SGD(optim.SGD):
    def __init__(self, lr: float = 1e-3, momentum: float = 0.9, dampening: float = 0.0,
                 weight_decay: float = 1e-8, nesterov: bool = False, *,
                 params: Iterable[nn.Parameter], **_) -> None:
        super(SGD, self).__init__(
            params=params, lr=lr,
            momentum=momentum, dampening=dampening,
            weight_decay=weight_decay, nesterov=nesterov,
        )


class Adam(optim.AdamW):
    def __init__(self, lr: float = 3e-4, beta1: float = 0.9, beta2: float = 0.98,
                 weight_decay: float = 1e-8, amsgrad: bool = False, *,
                 params: Iterable[nn.Parameter], **_) -> None:
        super(Adam, self).__init__(
            params=params, lr=lr, betas=(beta1, beta2),
            weight_decay=weight_decay, amsgrad=amsgrad,
        )



from abc import ABCMeta
from collections import Counter
from typing import Optional, Union, List, Any, Tuple

from torchglyph.proc import Proc, Processors, compress, subs, Identity
from torchglyph.vocab import Vocab

__all__ = [
    'Pipe', 'RawPipe',
]


class Pipe(object, metaclass=ABCMeta):
    def __init__(self, pre: Processors = None, vocab: Processors = None,
                 post: Processors = None, batch: Processors = None) -> None:
        super(Pipe, self).__init__()

        self.vocab: Optional[Union[Vocab]] = None

        self.pre_proc = Proc.from_list(compress(processors=pre))
        self.vocab_proc = Proc.from_list(compress(processors=vocab))
        self.post_proc = Proc.from_list(compress(processors=post))
        self.batch_proc = Proc.from_list(compress(processors=batch))

    def with_(self, pre: Processors = ..., vocab: Processors = ...,
              post: Processors = ..., batch: Processors = ...) -> 'Pipe':
        self.pre_proc = Proc.from_list(subs(processors=pre, repl=self.pre_proc))
        self.vocab_proc = Proc.from_list(subs(processors=vocab, repl=self.vocab_proc))
        self.post_proc = Proc.from_list(subs(processors=post, repl=self.post_proc))
        self.batch_proc = Proc.from_list(subs(processors=batch, repl=self.batch_proc))
        return self

    def extra_repr(self) -> str:
        return ',\n  '.join([
            f'pre={self.pre_proc}',
            f'vocab={self.vocab_proc}',
            f'post={self.post_proc}',
            f'batch={self.batch_proc}',
        ])

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(\n  {self.extra_repr()}\n)'

    def preprocess_(self, *datasets, counter: Optional[Counter] = None) -> Counter:
        if counter is None:
            counter = Counter()

        if not isinstance(self.pre_proc, Identity):
            for dataset in datasets:
                for name, pipe in dataset.pipes.items():
                    if self is pipe:
                        todo = f'{name}_pre_todo'
                        if getattr(dataset, todo, True):
                            dataset.data[name] = [
                                self.pre_proc(datum, counter=counter, name=name)
                                for datum in dataset.data[name]
                            ]
                            setattr(dataset, todo, False)

        return counter

    def postprocess_(self, *datasets) -> 'Pipe':
        _ = self.preprocess_(*datasets)

        if not isinstance(self.post_proc, Identity):
            for dataset in datasets:
                for name, pipe in dataset.pipes.items():
                    if self is pipe:
                        todo = f'{name}_post_todo'
                        if getattr(dataset, todo, True):
                            dataset.data[name] = [
                                self.post_proc(datum, vocab=self.vocab, name=name)
                                for datum in dataset.data[name]
                            ]
                            setattr(dataset, todo, False)

        return self

    def build_vocab_(self, *datasets, special_tokens: Tuple[str, ...] = (),
                     max_size: Optional[int] = None, min_freq: int = 1) -> 'Pipe':
        name = ', '.join(sorted(list(set([
            name for dataset in datasets
            for name, pipe in dataset.pipes.items() if self is pipe
        ]))))

        self.vocab = self.vocab_proc(
            self.preprocess_(*datasets),
            name=f'[{name}]' if ', ' in name else name,
            special_tokens=special_tokens,
            max_size=max_size, min_freq=min_freq,
        )

        return self

    def __call__(self, data: List[Any], name: str = '__call__') -> Tuple[Any, Vocab]:
        counter = Counter()
        data = [self.pre_proc(datum, counter=counter, name=name) for datum in data]
        vocab = self.vocab_proc(counter, name=name, special_tokens=(), max_size=None, min_freq=1)
        batch = [self.post_proc(datum, vocab=vocab, name=name) for datum in data]
        return self.batch_proc(batch, vocab=vocab, name=name), vocab

    def collate_fn(self, batch: List[Any]) -> Any:
        return self.batch_proc(batch, vocab=self.vocab)


class RawPipe(Pipe):
    def __init__(self) -> None:
        super(RawPipe, self).__init__(
            pre=None,
            vocab=None,
            post=None,
            batch=None,
        )



from abc import ABCMeta
from typing import List, Any

import torch
from torch import Tensor
from torch.nn.utils.rnn import PackedSequence
from torch.types import Device
from torchrua import accumulate_sizes, pack_padded_sequence, CattedSequence, PaddedSequence
from torchrua import pack_catted_sequence, pack_sequence
from torchrua import reduce_catted_sequences, batch_sizes_to_ptr

from torchglyph.proc.abc import Proc

__all__ = [
    'PackingProc',
    'PackSequence',
    'PackCattedSequence',
    'PackPaddedSequence',
    'ToPackedPtrSequence',
    'ReduceCattedSequences',
]


class PackingProc(Proc, metaclass=ABCMeta):
    def __init__(self, device: Device = None) -> None:
        super(PackingProc, self).__init__()
        self.device = device

    def extra_repr(self) -> str:
        return f'device={self.device}'

    def __call__(self, data: Any, **kwargs) -> PackedSequence:
        raise NotImplementedError


class PackSequence(PackingProc):
    def __call__(self, data: List[Tensor], **kwargs) -> PackedSequence:
        return pack_sequence(sequences=data, device=self.device)


class PackCattedSequence(PackingProc):
    def __call__(self, data: List[CattedSequence], **kwargs) -> PackedSequence:
        data, token_sizes = data

        if self.device is not None:
            data = data.to(device=self.device)
            token_sizes = token_sizes.to(device=self.device)
        return pack_catted_sequence(sequence=data, token_sizes=token_sizes, device=self.device)


class PackPaddedSequence(PackingProc):
    def __init__(self, batch_first: bool = True, device: Device = None) -> None:
        super(PackPaddedSequence, self).__init__(device=device)
        self.batch_first = batch_first

    def extra_repr(self) -> str:
        return ', '.join([
            f'batch_first={self.batch_first}',
            super(PackPaddedSequence, self).extra_repr()
        ])

    def __call__(self, data: List[PaddedSequence], **kwargs) -> PackedSequence:
        data, token_sizes = data

        return pack_padded_sequence(
            sequence=data.to(device=self.device),
            token_sizes=token_sizes.to(device=self.device),
            batch_first=self.batch_first, device=self.device,
        )


class ToPackedPtrSequence(PackingProc):
    @torch.no_grad()
    def __call__(self, data: PackedSequence, **kwargs) -> PackedSequence:
        device = self.device
        if device is None:
            device = data.data.device

        data = data.data.to(device=device)
        batch_sizes = data.batch_sizes.to(device=device)
        acc_batch_sizes = accumulate_sizes(sizes=batch_sizes)

        _, batch_ptr, _ = batch_sizes_to_ptr(batch_sizes=batch_sizes)

        return PackedSequence(
            data=acc_batch_sizes[data] + batch_ptr,
            batch_sizes=data.batch_sizes,
            sorted_indices=data.sorted_indices,
            unsorted_indices=data.unsorted_indices,
        )


class ReduceCattedSequences(PackingProc):
    def __call__(self, data: List[CattedSequence], **kwargs) -> PackedSequence:
        return reduce_catted_sequences(sequences=data, device=self.device)



from abc import ABCMeta
from typing import List, Any

from torch import Tensor
from torch.nn.utils.rnn import PackedSequence
from torch.types import Device
from torchrua import CattedSequence, cat_sequence, cat_packed_sequence, cat_padded_sequence

from torchglyph.proc.abc import Proc

__all__ = [
    'CattingProc',
    'CatSequence',
    'CatPackedSequence',
    'CatPaddedSequence',
]


class CattingProc(Proc, metaclass=ABCMeta):
    def __init__(self, device: Device = None) -> None:
        super(CattingProc, self).__init__()
        self.device = device

    def extra_repr(self) -> str:
        return f'device={self.device}'

    def __call__(self, data: Any, **kwargs) -> CattedSequence:
        raise NotImplementedError


class CatSequence(CattingProc):
    def __call__(self, data: List[Tensor], **kwargs) -> CattedSequence:
        return cat_sequence(sequences=data, device=self.device)


class CatPackedSequence(CattingProc):
    def __call__(self, data: PackedSequence, **kwargs) -> CattedSequence:
        return cat_packed_sequence(sequence=data, device=self.device)


class CatPaddedSequence(CattingProc):
    def __init__(self, batch_first: bool = True, device: Device = None) -> None:
        super(CatPaddedSequence, self).__init__(device=device)
        self.batch_first = batch_first

    def extra_repr(self) -> str:
        return ', '.join([
            f'batch_first={self.batch_first}',
            super(CatPaddedSequence, self).extra_repr()
        ])

    def __call__(self, data: PackedSequence, **kwargs) -> CattedSequence:
        data, token_sizes = data
        return cat_padded_sequence(
            sequence=data, token_sizes=token_sizes,
            batch_first=self.batch_first, device=self.device,
        )



from logging import getLogger

from torch.optim import Optimizer
from torch.optim.lr_scheduler import LambdaLR

__all__ = [
    'SchedulerMixin',
    'Fixed', 'InverseDecay',
]

logger = getLogger(__name__)


class SchedulerMixin(object):
    def __init__(self) -> None:
        self.batch = 1
        self.epoch = 1

    def batch_step(self) -> None:
        self.batch += 1

    def epoch_step(self) -> None:
        self.epoch += 1

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.extra_repr()})'

    def extra_repr(self) -> str:
        return ''

    def report_lr(self) -> None:
        lr = ', '.join(f'{lr:.6f}' for lr in self.get_lr())
        logger.info(f"learning rate => [{lr}]")


class InverseDecay(LambdaLR, SchedulerMixin):
    def __init__(self, gamma: float = 0.05, *, optimizer: Optimizer, **_) -> None:
        super(InverseDecay, self).__init__(
            optimizer=optimizer,
            lr_lambda=lambda epoch: 1. / (1. + gamma * epoch)
        )
        self.gamma = gamma

    def extra_repr(self) -> str:
        return f'gamma={self.gamma}'

    def epoch_step(self) -> None:
        self.step()
        super(InverseDecay, self).epoch_step()
        self.report_lr()


class Fixed(InverseDecay):
    def __init__(self, *, optimizer: Optimizer, **_) -> None:
        super(Fixed, self).__init__(gamma=0., optimizer=optimizer)



import logging
from collections import Counter, OrderedDict
from pathlib import Path
from typing import Optional, Tuple, List, Dict

import torch
from torch import Tensor
from torch.nn import init
from tqdm import tqdm

from torchglyph import data_dir
from torchglyph.io import DownloadMixin

logger = logging.getLogger(__name__)

__all__ = [
    'Vocab', 'Vectors',
    'Glove', 'FastText',
]


class Vocab(object):
    def __init__(self, counter: Counter,
                 unk_token: Optional[str],
                 pad_token: Optional[str],
                 special_tokens: Tuple[Optional[str], ...] = (),
                 max_size: Optional[int] = None, min_freq: int = 1) -> None:
        super(Vocab, self).__init__()

        self.freq = counter
        self.itos: Dict[int, str] = {}
        self.stoi: Dict[str, int] = {}
        self.vectors: Optional[Tensor] = None

        self.unk_idx = None
        if unk_token is not None:
            self.unk_idx = self.add_token_(unk_token)

        self.pad_idx = None
        if pad_token is not None:
            self.pad_idx = self.add_token_(pad_token)

        for token in special_tokens:
            self.add_token_(token)

        self.unk_token = unk_token
        self.pad_token = pad_token
        self.special_tokens = tuple(self.stoi.keys())

        for token, freq in self.freq.most_common(n=max_size):
            if freq < min_freq:
                break
            self.add_token_(token)

        assert len(self.stoi) == len(self.itos)

    def __getitem__(self, token: str) -> int:
        return self.stoi.get(token, self.unk_idx)

    def inv(self, index: int) -> str:
        return self.itos.get(index, self.unk_token)

    def add_token_(self, token: str) -> int:
        assert token is not None

        if token not in self.stoi:
            index = len(self.stoi)
            self.stoi[token] = index
            self.itos[index] = token

        return self.stoi[token]

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.extra_repr()})'

    def extra_repr(self) -> str:
        return ', '.join([
            f'{len(self.stoi)}',
            *self.special_tokens,
        ])

    def __len__(self) -> int:
        return len(self.stoi)

    def __contains__(self, token: str) -> bool:
        return token in self.stoi

    @torch.no_grad()
    def load_vectors(self, *fallbacks, vectors: 'Vectors') -> Tuple[int, int]:
        _, vector_size = vectors.vectors.size()
        self.vectors = torch.empty((len(self), vector_size), dtype=torch.float32)

        tok, occ = 0, 0
        for token, index in self.stoi.items():
            if vectors.query_(token, self.vectors[index], *fallbacks):
                tok += 1
                occ += self.freq[token]

        if self.pad_token is not None:
            init.zeros_(self.vectors[self.stoi[self.pad_token]])

        return tok, occ

    def state_dict(self, destination: OrderedDict = None, prefix: str = '', **kwargs) -> OrderedDict:
        if destination is None:
            destination = OrderedDict()
            destination._metadata = OrderedDict()

        destination[prefix + 'unk_idx'] = self.unk_idx
        destination[prefix + 'pad_idx'] = self.pad_idx
        destination[prefix + 'unk_token'] = self.unk_token
        destination[prefix + 'pad_token'] = self.pad_token
        destination[prefix + 'special_tokens'] = self.special_tokens

        destination[prefix + 'freq'] = self.freq
        destination[prefix + 'stoi'] = self.stoi
        destination[prefix + 'itos'] = self.itos
        destination[prefix + 'vectors'] = self.vectors.detach()

        return destination

    def load_state_dict(self, state_dict: OrderedDict, strict: bool = True, **kwargs) -> None:
        self.unk_idx = state_dict.pop('unk_idx')
        self.pad_idx = state_dict.pop('pad_idx')
        self.unk_token = state_dict.pop('unk_token')
        self.pad_token = state_dict.pop('pad_token')
        self.special_tokens = state_dict.pop('special_tokens')

        self.freq = state_dict.pop('freq')
        self.stoi = state_dict.pop('stoi')
        self.itos = state_dict.pop('itos')
        self.vectors = state_dict.pop('vectors')

        if strict:
            assert len(state_dict) == 0


class Vectors(Vocab, DownloadMixin):
    vector_format: str

    def __init__(self, root: Path = data_dir, **kwargs) -> None:
        super(Vectors, self).__init__(
            counter=Counter(),
            unk_token=None,
            pad_token=None,
            special_tokens=(),
            max_size=None, min_freq=1,
        )

        path, = self.paths(root=root, **kwargs)  # type:Path
        self.cache_(path=path)

    def cache_(self, path: Path) -> None:
        torch_path = path.with_suffix('.pt')

        if torch_path.exists():
            logger.info(f'loading from {torch_path}')
            self.load_state_dict(state_dict=torch.load(f=torch_path))
        else:
            vectors = []
            with path.open('r', encoding='utf-8') as fp:

                num_embeddings, embedding_dim = None, None
                if self.vector_format == 'word2vec':
                    num_embeddings, embedding_dim = map(int, next(fp).strip().split(' '))

                for raw in tqdm(fp, desc=f'caching {path}', unit=' tokens', total=num_embeddings):
                    token, *embeddings = raw.rstrip().split(' ')

                    if embedding_dim is None:
                        embedding_dim = len(embeddings)
                    assert embedding_dim == len(embeddings), f'{embedding_dim} != {len(embeddings)} :: {token}'

                    self.add_token_(token)
                    vectors.append([float(v) for v in embeddings])

            self.vectors = torch.tensor(vectors, dtype=torch.float32)

            logger.info(f'saving to {torch_path}')
            torch.save(obj=self.state_dict(), f=torch_path)

    @torch.no_grad()
    def query_(self, token: str, tensor: Tensor, *fallbacks) -> bool:
        if token in self:
            tensor[:] = self.vectors[self.stoi[token]]
            return True

        for fallback in fallbacks:
            new_token = fallback(token)
            if new_token in self:
                tensor[:] = self.vectors[self.stoi[new_token]]
                return True

        self.unk_init_(tensor)
        return False

    @staticmethod
    def unk_init_(tensor: Tensor) -> None:
        init.normal_(tensor)


class Glove(Vectors):
    vector_format = 'glove'

    def __init__(self, name: str, dim: int, root: Path = data_dir) -> None:
        super(Glove, self).__init__(root=root, name=name, dim=dim)

    @classmethod
    def get_urls(cls, name: str, dim: int) -> List[Tuple[str, ...]]:
        return [(
            f'https://nlp.stanford.edu/data/glove.{name}.zip',
            f'glove.{name}.zip',
            f'glove.{name}.{dim}d.txt',
        )]


class FastText(Vectors):
    vector_format = 'word2vec'

    def __init__(self, name: str, lang: str, root: Path = data_dir) -> None:
        super(FastText, self).__init__(root=root, name=name, lang=lang)

    @classmethod
    def get_urls(cls, name: str, lang: str) -> List[Tuple[str, ...]]:
        if name == 'cc':
            return [(
                f'https://dl.fbaipublicfiles.com/fasttext/vectors-crawl/cc.{lang}.300.vec.gz',
                f'cc.{lang}.300.vec.gz',
                f'cc.{lang}.300.vec',
            )]
        if name == 'wiki':
            return [(
                f'https://dl.fbaipublicfiles.com/fasttext/vectors-wiki/wiki.{lang}.vec',
                f'wiki.{lang}.vec',
                f'wiki.{lang}.vec',
            )]
        if name == 'aligned':
            return [(
                f'https://dl.fbaipublicfiles.com/fasttext/vectors-aligned/wiki.{lang}.align.vec',
                f'wiki.{lang}.align.vec',
                f'wiki.{lang}.align.vec',
            )]

        raise KeyError(f'{name} is not supported')


if __name__ == '__main__':
    vectors = FastText(name='cc', lang='en')
    print(f'vectors => {vectors}')
    print(vectors.vectors.size())



from typing import Iterable, List, Tuple, IO

from tqdm import tqdm

__all__ = [
    'load_meta',
    'loads_vector',
    'load_vector',
    'iter_vector',
    'load_word2vec', 'load_glove',
]

Token = str
Vector = Tuple[float, ...]


def load_meta(fp: IO, *, sep: str = ' ') -> Tuple[int, int]:
    num_embeddings, embedding_dim = next(fp).strip().split(sep=sep)
    return int(num_embeddings), int(embedding_dim)


def loads_vector(s: str, *, sep: str = ' ') -> Tuple[Token, Vector]:
    token, *scalars = s.strip().split(sep=sep)
    return str(token), tuple(float(scalar) for scalar in scalars)


def load_vector(fp: IO, *, sep: str = ' ') -> Tuple[List[Token], List[Vector]]:
    tokens, vectors = zip(*[loads_vector(s, sep=sep) for s in fp])
    return tokens, vectors


def iter_vector(fp: IO, *, sep: str = ' ') -> Iterable[Tuple[Token, Vector]]:
    yield from map(lambda s: loads_vector(s, sep=sep), fp)


def load_word2vec(fp: IO, *, sep: str = ' ') -> Tuple[List[Token], List[Vector]]:
    num_embeddings, embedding_dim = load_meta(fp, sep=sep)
    if isinstance(fp, tqdm) and hasattr(fp, 'total'):
        fp.total = num_embeddings

    tokens, vectors = [], []
    for token, vector in iter_vector(fp, sep=sep):
        assert len(vector) == embedding_dim, f'len({token}) = {len(vector)} != {embedding_dim}'

        tokens.append(token)
        vectors.append(vector)

    return tokens, vectors


def load_glove(fp: IO, *, sep: str = ' ') -> Tuple[List[Token], List[Vector]]:
    token, vector = loads_vector(next(fp), sep=sep)
    embedding_dim = len(vector)

    tokens, vectors = [token], [vector]
    for token, vector in iter_vector(fp, sep=sep):
        assert len(vector) == embedding_dim, f'len({token}) = {len(vector)} != {embedding_dim}'

        tokens.append(token)
        vectors.append(vector)

    return tokens, vectors



import torch
from torch.types import Device
from transformers import AutoTokenizer

from torchglyph.pipe import Pipe
from torchglyph.proc.ctx import CtxTokenize, CtxCollate

__all__ = [
    'CtxPipe',
]


class CtxPipe(Pipe):
    def __init__(self, pretrained_model_name: str, device: Device, add_special_tokens: bool = True,
                 prefix: str = '##', dtype: torch.dtype = torch.long) -> None:
        tokenizer = AutoTokenizer.from_pretrained(pretrained_model_name)

        super(CtxPipe, self).__init__(
            pre=CtxTokenize(tokenizer=tokenizer, add_special_tokens=add_special_tokens, prefix=prefix, dtype=dtype),
            vocab=None,
            post=None,
            batch=CtxCollate(tokenizer=tokenizer, device=device),
        )



from typing import Any, Type, Tuple, NamedTuple, IO, Iterable, get_type_hints

from torchglyph.formats.primitive import loads_type, dumps_type

__all__ = [
    'loads_token', 'iter_sentence',
    'dumps_token', 'dump_sentence',
]

Token = Tuple[Any, ...]
Sentence = Tuple[Tuple[Any, ...], ...]


def loads_token(s: str, *, config: Type[NamedTuple], sep: str = '\t') -> Token:
    return tuple(
        loads_type(s, tp=tp)
        for s, (name, tp) in zip(s.strip().split(sep=sep), get_type_hints(config).items())
        if not name.endswith('_')
    )


def iter_sentence(fp: IO, *, config: Type[NamedTuple], sep: str = '\t', blank: str = '') -> Iterable[Sentence]:
    sentence = []

    for s in fp:
        s = s.strip()
        if s != blank:
            sentence.append(loads_token(s, config=config, sep=sep))
        elif len(sentence) != 0:
            yield tuple(zip(*sentence))
            sentence = []

    if len(sentence) != 0:
        yield tuple(zip(*sentence))


def dumps_token(token: Token, sep: str = '\t') -> str:
    return sep.join(map(dumps_type, token))


def dump_sentence(sentences: Iterable[Sentence], fp: IO, *, sep: str = '\t', blank: str = '') -> None:
    for tokens in sentences:
        for token in tokens:
            fp.write(dumps_token(token, sep=sep))
            fp.write('\n')

        fp.write(blank)
        fp.write('\n')



import torch
from torch import Tensor
from torch import nn
from torch.nn import functional as F

from torchglyph.nn import init

__all__ = [
    'Linear', 'ConjugatedLinear',
    'Projector', 'ConjugatedProjector',
]


class Linear(nn.Linear):
    def __init__(self, bias: bool = True, normalize: bool = False, *,
                 in_features: int, out_features: int,
                 dtype: torch.dtype = torch.float32) -> None:
        super(Linear, self).__init__(
            in_features=in_features, out_features=out_features,
            bias=bias, dtype=dtype,
        )
        self.normalize = normalize

    def reset_parameters(self) -> None:
        init.kaiming_uniform_(self.weight, fan=self.in_features, a=5. ** 0.5, nonlinearity='leaky_relu')
        if self.bias is not None:
            init.constant_(self.bias, 0.)

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        if self.normalize:
            cls_name = f'Cos{cls_name}'
        return f'{cls_name}({self.extra_repr()})'

    def forward(self, tensor: Tensor) -> Tensor:
        if not self.normalize:
            weight = self.weight
        else:
            tensor = F.normalize(tensor, p=2, dim=-1)
            weight = F.normalize(self.weight, p=2, dim=-1)

        return F.linear(tensor, weight=weight, bias=self.bias)


class ConjugatedLinear(nn.Module):
    def __init__(self, bias: bool = True, normalize: bool = False, *,
                 in_features: int, num_conjugates: int, out_features: int,
                 dtype: torch.dtype = torch.float32) -> None:
        super(ConjugatedLinear, self).__init__()

        self.weight = nn.Parameter(torch.empty((num_conjugates, out_features, in_features), dtype=dtype))
        if bias:
            self.bias = nn.Parameter(torch.empty((num_conjugates, out_features), dtype=dtype))
        else:
            self.register_parameter('bias', None)

        self.in_features = in_features
        self.num_conjugates = num_conjugates
        self.out_features = out_features
        self.normalize = normalize

        self.reset_parameters()

    def reset_parameters(self) -> None:
        init.kaiming_uniform_(self.weight, fan=self.in_features, a=5. ** 0.5, nonlinearity='leaky_relu')
        if self.bias is not None:
            init.constant_(self.bias, 0.)

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        if self.normalize:
            cls_name = f'Cos{cls_name}'
        return f'{cls_name}({self.extra_repr()})'

    def extra_repr(self) -> str:
        return ', '.join([
            f'in_features={self.in_features}',
            f'num_conjugates={self.num_conjugates}',
            f'out_features={self.out_features}',
            f'bias={self.bias is not None}',
        ])

    def forward(self, tensor: Tensor) -> Tensor:
        if not self.normalize:
            weight = self.weight
        else:
            tensor = F.normalize(tensor, p=2, dim=-1)
            weight = F.normalize(self.weight, p=2, dim=-1)

        out = (weight @ tensor[..., None])[..., 0]
        if self.bias is not None:
            out = out + self.bias
        return out


class Projector(Linear):
    def __init__(self, normalize: bool = False, *,
                 in_features: int, out_features: int,
                 dtype: torch.dtype = torch.float32) -> None:
        super(Projector, self).__init__(
            in_features=in_features, out_features=out_features,
            bias=False, normalize=normalize, dtype=dtype,
        )

    def reset_parameters(self) -> None:
        init.kaiming_uniform_(self.weight, fan=self.in_features, a=0., nonlinearity='relu')
        if self.bias is not None:
            init.constant_(self.bias, 0.)


class ConjugatedProjector(ConjugatedLinear):
    def __init__(self, normalize: bool = False, *,
                 in_features: int, num_conjugates: int, out_features: int,
                 dtype: torch.dtype = torch.float32) -> None:
        super(ConjugatedProjector, self).__init__(
            in_features=in_features, num_conjugates=num_conjugates, out_features=out_features,
            bias=False, normalize=normalize, dtype=dtype,
        )

    def reset_parameters(self) -> None:
        init.kaiming_uniform_(self.weight, fan=self.in_features, a=0., nonlinearity='relu')
        if self.bias is not None:
            init.constant_(self.bias, 0.)



import logging
from pathlib import Path
from typing import Iterable, NamedTuple, Any, List

import torch
from torch.types import Device
from tqdm import tqdm

from torchglyph import data_dir
from torchglyph.dataset import Dataset, DataLoader
from torchglyph.formats.conll import iter_sentence
from torchglyph.pipe.packing import PackListStrPipe, PackListListStrPipe

__all__ = [
    'CoNLL2003',
]


class WordPipe(PackListStrPipe):
    def __init__(self, device: Device) -> None:
        super(WordPipe, self).__init__(
            device=device, dtype=torch.long,
            unk_token='<unk>', special_tokens=(), threshold=10,
        )


class CharPipe(PackListListStrPipe):
    def __init__(self, device: Device) -> None:
        super(CharPipe, self).__init__(
            device=device, dtype=torch.long,
            unk_token='<unk>', special_tokens=(), threshold=10,
        )


class TagPipe(PackListStrPipe):
    def __init__(self, device: Device) -> None:
        super(TagPipe, self).__init__(
            device=device, dtype=torch.long,
            unk_token='O', special_tokens=(), threshold=120,
        )


class CoNLL2003(Dataset):
    urls = [
        ('https://raw.githubusercontent.com/glample/tagger/master/dataset/eng.train', 'train.txt'),
        ('https://raw.githubusercontent.com/glample/tagger/master/dataset/eng.testa', 'dev.txt'),
        ('https://raw.githubusercontent.com/glample/tagger/master/dataset/eng.testb', 'test.txt'),
    ]

    class Config(NamedTuple):
        word: str
        pos_: str
        chunk_: str
        tag: str

    def get_size(self, item: Any) -> int:
        return item['word'].size()[0]

    @classmethod
    def load(cls, path: Path, **kwargs) -> Iterable[NamedTuple]:
        with tqdm(path.open(mode='r', encoding='Utf-8'), desc=f'loading {path}') as fp:
            for item in iter_sentence(fp, config=cls.Config, sep=' '):
                yield item

    @classmethod
    def new(cls, batch_size: int, *, device: Device,
            root: Path = data_dir, **kwargs) -> List['DataLoader']:
        word = WordPipe(device=device)
        char = CharPipe(device=device)
        tag = TagPipe(device=device)

        pipes = [
            dict(word=word, char=char),
            dict(tag=tag),
        ]

        for ps in pipes:
            for name, pipe in ps.items():
                logging.info(f'{name} => {pipe}')

        train, dev, test = cls.paths(root=root)
        train = cls(pipes=pipes, path=train)
        dev = cls(pipes=pipes, path=dev)
        test = cls(pipes=pipes, path=test)

        word.build_vocab_(train)
        char.build_vocab_(train)
        tag.build_vocab_(train)

        return DataLoader.new(
            (train, dev, test),
            batch_size=batch_size,
            shuffle=True, drop_last=False,
        )



import heapq
import logging
from collections import Counter
from typing import Tuple, Optional

from torchglyph.proc.abc import Proc, Map
from torchglyph.vocab import Vocab, Vectors, Glove, FastText

logger = logging.getLogger(__name__)

__all__ = [
    'UpdateCounter', 'Numbering', 'BuildVocab', 'StatsVocab',
    'LoadVectors', 'LoadGlove', 'LoadFastText',
]


class UpdateCounter(Map):
    def map(self, token: str, *, counter: Counter, **kwargs) -> str:
        counter[token] += 1
        return token


class Numbering(Map):
    def map(self, token: str, *, vocab: Vocab, **kwargs) -> int:
        return vocab[token]


class BuildVocab(Proc):
    def __init__(self, unk_token: Optional[str], pad_token: Optional[str],
                 special_tokens: Tuple[Optional[str], ...] = ()) -> None:
        super(BuildVocab, self).__init__()
        self.unk_token = unk_token
        self.pad_token = pad_token
        special_tokens = (unk_token, pad_token, *special_tokens)
        self.special_tokens = tuple(token for token in special_tokens if token is not None)

    def extra_repr(self) -> str:
        return ', '.join(self.special_tokens)

    def __call__(self, vocab: Counter, *, max_size: Optional[int], min_freq: int, **kwargs) -> Vocab:
        return Vocab(
            counter=vocab,
            unk_token=self.unk_token,
            pad_token=self.pad_token,
            special_tokens=self.special_tokens,
            max_size=max_size, min_freq=min_freq,
        )


class StatsVocab(Proc):
    def __init__(self, threshold: int) -> None:
        super(StatsVocab, self).__init__()
        self.threshold = threshold

    def extra_repr(self) -> str:
        return f'{self.threshold}'

    def __call__(self, vocab: Vocab, name: str, *args, **kwargs) -> Vocab:
        num_tokens = len(vocab.freq)
        avg_freq = sum(vocab.freq.values()) / max(1, num_tokens)

        if num_tokens <= self.threshold:
            xs = vocab.freq.most_common()
            xs = ', '.join([f"'{token}'({freq})" for token, freq in xs])

            logger.info(f"{name}.vocab => {vocab} :: {avg_freq:.1f} times/token")
            logger.info(f"{name}.tokens => [{xs}]")

        else:
            n, m = self.threshold // 2, (self.threshold + 1) // 2
            xs = heapq.nlargest(n, vocab.freq.items(), key=lambda x: x[1])
            ys = heapq.nsmallest(m, vocab.freq.items(), key=lambda x: x[1])
            xs = ', '.join([f"'{token}'({freq})" for token, freq in xs])
            ys = ', '.join([f"'{token}'({freq})" for token, freq in ys[::-1]])

            logger.info(f"{name}.vocab => {vocab} :: {avg_freq:.1f} times/token")
            logger.info(f"{name}.tokens => [{xs}, ..., {ys}]")

        return vocab


class LoadVectors(Proc):
    def __init__(self, *fallbacks, vectors: Vectors) -> None:
        super(LoadVectors, self).__init__()
        self.fallbacks = fallbacks
        self.vectors = vectors

    def extra_repr(self) -> str:
        return ', '.join([
            self.vectors.__class__.__name__,
            *[f'{fallback.__name__}' for fallback in self.fallbacks],
        ])

    def __call__(self, vocab: Vocab, *, name: str, **kwargs) -> Vocab:
        assert vocab is not None, \
            f"did you forget '{BuildVocab.__name__}' before '{LoadVectors.__name__}'?"

        tok, occ = vocab.load_vectors(*self.fallbacks, vectors=self.vectors)
        tok = tok / max(1, len(vocab.freq.values())) * 100
        occ = occ / max(1, sum(vocab.freq.values())) * 100

        logger.info(f"{self.vectors} hits {tok:.1f}% tokens "
                    f"and {occ:.1f}% occurrences of {Vocab.__name__} '{name}'")
        return vocab


class LoadGlove(LoadVectors):
    def __init__(self, *fallbacks, name: str, dim: int) -> None:
        super(LoadGlove, self).__init__(
            *fallbacks, vectors=Glove(name=name, dim=dim),
        )


class LoadFastText(LoadVectors):
    def __init__(self, *fallbacks, name: str, lang: str) -> None:
        super(LoadFastText, self).__init__(
            *fallbacks, vectors=FastText(name=name, lang=lang),
        )



import re
from typing import Pattern, Any, List, Tuple, Set, Union, Sized

from torchglyph.proc.abc import Proc, Map

__all__ = [
    'ToStr', 'ToInt', 'ToBool', 'ToFloat',
    'ToSet', 'ToList', 'ToTuple', 'ToSize',
    'ToLower', 'ToUpper', 'ToCapitalized', 'RegexSub',
    'Prepend', 'Append',
]


class ToStr(Map):
    def map(self, data: Any, **kwargs) -> str:
        return str(data)


class ToInt(Map):
    def map(self, data: Any, **kwargs) -> int:
        return int(data)


class ToBool(Map):
    def map(self, data: Any, **kwargs) -> bool:
        return bool(data)


class ToFloat(Map):
    def map(self, data: Any, **kwargs) -> float:
        return float(data)


class ToSet(Proc):
    def __call__(self, data: Any, **kwargs) -> Set[Any]:
        return set(data)


class ToList(Proc):
    def __call__(self, data: Any, **kwargs) -> List[Any]:
        return list(data)


class ToTuple(Proc):
    def __call__(self, data: Any, **kwargs) -> Tuple[Any, ...]:
        return tuple(data)


class ToSize(Proc):
    def __call__(self, data: Sized, **kwargs) -> int:
        return len(data)


class ToLower(Map):
    def map(self, string: str, **kwargs) -> str:
        return string.lower()


class ToUpper(Map):
    def map(self, string: str, **kwargs) -> str:
        return string.upper()


class ToCapitalized(Map):
    def map(self, string: str, **kwargs) -> str:
        return string.capitalize()


class RegexSub(Map):
    def __init__(self, pattern: Pattern, repl: str) -> None:
        super(RegexSub, self).__init__()
        self.pattern = pattern
        self.repl = repl

    def extra_repr(self) -> str:
        return f'{self.pattern} -> {self.repl}'

    def map(self, string: str, **kwargs) -> str:
        return re.sub(pattern=self.pattern, repl=self.repl, string=string)


class Prepend(Proc):
    Container = Union[Set[Any], List[Any], Tuple[Any, ...]]

    def __init__(self, token: Any, num_times: int = 1) -> None:
        super(Prepend, self).__init__()
        self.token = token
        self.num_times = num_times

    def extra_repr(self) -> str:
        return f'{self.token}(x{self.num_times}) + ...'

    def __call__(self, sequence: Container, **kwargs) -> Container:
        return type(sequence)([self.token for _ in range(self.num_times)] + sequence)


class Append(Proc):
    Container = Union[Set[Any], List[Any], Tuple[Any, ...]]

    def __init__(self, token: Any, num_times: int = 1) -> None:
        super(Append, self).__init__()
        self.token = token
        self.num_times = num_times

    def extra_repr(self) -> str:
        return f'... + {self.token}(x{self.num_times})'

    def __call__(self, sequence: Container, **kwargs) -> Container:
        return type(sequence)(sequence + [self.token for _ in range(self.num_times)])



from typing import Any, List

import torch
from torch import Tensor
from torch.types import Device
from torchrua import cat_sequence, token_sizes_to_mask, pad_catted_sequence
from transformers import PreTrainedTokenizer

from torchglyph.proc.abc import Proc

__all__ = [
    'CtxTokenize',
    'CtxCollate',
]


class CtxTokenize(Proc):
    def __init__(self, tokenizer: PreTrainedTokenizer,
                 add_special_tokens: bool, prefix: str,
                 dtype: torch.dtype) -> None:
        super(CtxTokenize, self).__init__()

        self.tokenizer = tokenizer
        self.add_special_tokens = add_special_tokens
        self.prefix = prefix
        self.dtype = dtype

    def extra_repr(self) -> str:
        return ', '.join([
            f'{self.tokenizer.name_or_path}',
            f'add_special_tokens={self.add_special_tokens}',
            f'prefix={self.prefix}',
        ])

    def __call__(self, text: str, **kwargs) -> Any:
        tokens = self.tokenizer.tokenize(text=text, add_special_tokens=self.add_special_tokens)
        ids = self.tokenizer.convert_tokens_to_ids(tokens=tokens)

        return torch.tensor([
            -id if token.startswith(self.prefix) else id
            for token, id in zip(tokens, ids)
        ], dtype=self.dtype)


class CtxCollate(Proc):
    def __init__(self, tokenizer: PreTrainedTokenizer, device: Device) -> None:
        super(CtxCollate, self).__init__()

        self.tokenizer = tokenizer
        self.device = device

    def extra_repr(self) -> str:
        return f'device={self.device}'

    def __call__(self, sequences: List[Tensor], **kwargs):
        sequence, token_sizes = cat_sequence(sequences, device=self.device)

        input_ids = pad_catted_sequence(
            sequence=sequence, token_sizes=token_sizes,
            batch_first=True, padding_value=self.tokenizer.pad_token_id,
        )
        attention_mask = token_sizes_to_mask(token_sizes=token_sizes, batch_first=True)

        return dict(
            input_ids=input_ids.abs(),
            attention_mask=attention_mask,
        )



from setuptools import setup, find_packages

name = 'torchglyph'

setup(
    name=name,
    version='0.3.0',
    packages=[package for package in find_packages() if package.startswith(name)],
    url=f'https://speedcell4.github.io/torchglyph',
    license='MIT',
    author='speedcell4',
    author_email='speedcell4@gmail.com',
    description='Data Processor Combinators for Natural Language Processing',
    install_requires=[
        'tqdm',
        'numpy',
        'einops',
        'torchrua>=0.3.0',
        'requests',
        'tabulate',
        'aku',
    ],
    extras_require={
        'dev': [
            'pytest',
            'hypothesis',
        ],
        'ctx': [
            'transformers',
        ],
        'docs': [
            'mkdocs',
            'mkdocs-alabaster',
        ]
    }
)



from typing import Any, List, Union, Set, Tuple

import numpy as np
import torch
from torch import Tensor
from torch.nn.utils.rnn import PackedSequence
from torch.types import Device

from torchglyph.proc.abc import Proc

__all__ = [
    'ToTensor',
    'ToDevice',
    'CatTensors',
    'StackTensors',
]


class ToTensor(Proc):
    def __init__(self, dtype: torch.dtype = None) -> None:
        super(ToTensor, self).__init__()
        self.dtype = dtype

    def extra_repr(self) -> str:
        if self.dtype is not None:
            return f'{self.dtype}'
        return ''

    def __call__(self, data: Any, **kwargs) -> Tensor:
        try:
            if torch.is_tensor(data):
                tensor = data.clone()
            elif isinstance(data, np.ndarray):
                tensor = torch.from_numpy(data).clone()
            else:
                tensor = torch.tensor(data)
            return tensor.to(dtype=self.dtype).requires_grad_(False)
        except ValueError as error:
            if error.args[0] == "too many dimensions 'str'":
                raise ValueError(f"'{data}' can not be converted to {Tensor.__name__}")
            raise error


class ToDevice(Proc):
    Tensors = Union[Tensor, PackedSequence, Set[Tensor], List[Tensor], Tuple[Tensor, ...]]

    def __init__(self, device: Device = None) -> None:
        super(ToDevice, self).__init__()
        self.device = device

    def extra_repr(self) -> str:
        return f'{self.device}'

    def __call__(self, tensors: Tensors, **kwargs) -> Tensors:
        if isinstance(tensors, (Tensor, PackedSequence)):
            return tensors.to(device=self.device)
        if isinstance(tensors, (set, list, tuple)):
            return type(tensors)([self(tensor, **kwargs) for tensor in tensors])
        return tensors


class CatTensors(Proc):
    def __init__(self, dim: int) -> None:
        super(CatTensors, self).__init__()
        self.dim = dim

    def extra_repr(self) -> str:
        return f'dim={self.dim}'

    def __call__(self, tensors: List[Tensor], **kwargs) -> Tensor:
        return torch.cat(tensors, dim=self.dim)


class StackTensors(Proc):
    def __init__(self, dim: int) -> None:
        super(StackTensors, self).__init__()
        self.dim = dim

    def extra_repr(self) -> str:
        return f'dim={self.dim}'

    def __call__(self, tensors: List[Tensor], **kwargs) -> Tensor:
        return torch.stack(tensors, dim=self.dim)



from typing import IO, Iterable, Tuple, Sequence, List

__all__ = [
    'load_freq', 'loads_freq', 'iter_freq',
    'dump_freq', 'dumps_freq',
]


def loads_freq(s: str, *, sep: str = '\t') -> Tuple[str, int]:
    token, freq = s.strip().split(sep=sep)
    return str(token), int(freq)


def load_freq(fp: IO, *, sep: str = '\t') -> List[Tuple[str, int]]:
    return [loads_freq(s, sep=sep) for s in fp]


def iter_freq(fp: IO, *, sep: str = '\t') -> Iterable[Tuple[str, int]]:
    yield from map(lambda s: loads_freq(s, sep=sep), fp)


def dumps_freq(token: str, freq: int, *, sep: str = '\t') -> str:
    return f'{token}{sep}{freq}'


def dump_freq(obj: Sequence[Tuple[str, int]], fp: IO, *, sep: str = '\t') -> None:
    for token, freq in obj:
        fp.write(dumps_freq(token, freq, sep=sep))
        fp.write('\n')



import gzip
import logging
import re
import shutil
import tarfile
import zipfile
from pathlib import Path
from typing import Union, Pattern, List, Tuple

import requests
from requests import Response
from tqdm import tqdm

from torchglyph import data_dir

logger = logging.getLogger(__name__)

__all__ = [
    'toggle_loggers',
    'DownloadMixin', 'download', 'extract',
]


def toggle_loggers(pattern: Union[str, Pattern], enable: bool) -> None:
    for name in logging.root.manager.loggerDict:  # type:str
        if re.match(pattern, name) is not None:
            logging.getLogger(name).disabled = not enable


def download(url: str, path: Path, exist_ok: bool = True, chunk_size: int = 1024 * 1024) -> Path:
    response: Response = requests.get(url=url, stream=True)
    assert response.status_code == 200, f'{response.status_code} != {200}'

    path.parent.mkdir(parents=True, exist_ok=True)
    if not path.exists() or not exist_ok:
        with path.open(mode='wb') as fp:
            for chunk in tqdm(response.iter_content(chunk_size=chunk_size),
                              total=int(response.headers['Content-Length']) // chunk_size,
                              desc=f'downloading from {url}', unit='MB'):
                fp.write(chunk)

    return path


def extract(path: Path) -> Path:
    logger.info(f'extracting files from {path}')

    if path.suffix == '.zip':
        logger.info(f'extracting {path}')
        with zipfile.ZipFile(path, "r") as fp:
            fp.extractall(path=path.parent)
    elif path.suffixes[-2:] == ['.tar', '.gz']:
        logger.info(f'extracting {path}')
        with tarfile.open(path, 'r:gz') as fp:
            fp.extractall(path=path.parent)
    elif path.suffix == '.gz':
        logger.info(f'extracting {path}')
        with gzip.open(path, mode='rb') as fs:
            with path.with_suffix('').open(mode='wb') as fd:
                shutil.copyfileobj(fs, fd)

    return path


class DownloadMixin(object):
    name: str
    urls: List[Tuple[str, ...]]

    @classmethod
    def get_urls(cls, **kwargs) -> List[Tuple[str, ...]]:
        return cls.urls

    @classmethod
    def paths(cls, root: Path = data_dir, **kwargs) -> List[Path]:
        root = root / getattr(cls, 'name', cls.__name__).lower()

        paths = []
        for url, path, *names in cls.get_urls(**kwargs):
            if len(names) == 0:
                names = [path]
            if any(not (root / name).exists() for name in names):
                extract(path=download(url=url, path=root / path, exist_ok=False))
            for name in names:
                if not (root / name).exists():
                    raise FileNotFoundError(f'{root / name} is not obtainable from {url}')
                paths.append(root / name)

        return paths



from typing import Type, Any

__all__ = [
    'loads_type', 'loads_bool',
    'dumps_type', 'dumps_bool',
]


def loads_bool(s: str) -> bool:
    s = s.strip().lower()
    if s in ('1', 'y', 'yes', 't', 'true'):
        return True
    if s in ('0', 'n', 'no', 'f', 'false'):
        return False
    raise TypeError(f'{s} is not a boolean value')


def loads_type(s: str, *, tp: Type) -> Any:
    if tp is bool:
        return loads_bool(s)
    return tp(s)


def dumps_bool(v: bool) -> str:
    return f'{v}'


def dumps_type(v: Any) -> str:
    if isinstance(v, bool):
        return dumps_bool(v)
    return f'{v}'



from typing import List, Iterator

from torch.utils.data import BatchSampler, Sampler

__all__ = [
    'SizedBatchSampler',
]


class SizedBatchSampler(BatchSampler):
    def __init__(self, dataset, sampler: Sampler[int], batch_size: int, drop_last: bool) -> None:
        super(SizedBatchSampler, self).__init__(sampler, batch_size, drop_last)

        self.dataset = dataset

        self.reset_indices()

    def reset_indices(self) -> None:
        self.indices, batch, batch_size = [], [], 0

        for index in self.sampler:
            data_size = self.dataset.get_size(self.dataset[index])
            batch.append(index)
            batch_size += data_size

            if batch_size >= self.batch_size:
                self.indices.append(batch)
                batch, batch_size = [], 0

        if not self.drop_last and batch_size > 0:
            self.indices.append(batch)

    def __iter__(self) -> Iterator[List[int]]:
        yield from self.indices
        self.reset_indices()

    def __len__(self) -> int:
        num_batches = len(self.indices)

        if not self.drop_last:
            num_batches += self.batch_size - 1
        return num_batches // self.batch_size
